// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file DuneSolver.h */

#ifndef AMDIS_DUNESOLVER_H
#define AMDIS_DUNESOLVER_H

#ifdef HAVE_DUNE

#define DUNE_MINIMAL_DEBUG_LEVEL 4

#include <dune/common/fmatrix.hh>
#include <dune/istl/bvector.hh>
#include <dune/istl/bcrsmatrix.hh>

#include <dune/istl/bvector.hh>
#include <dune/istl/bcrsmatrix.hh>
#include <dune/istl/ilu.hh>
#include <dune/istl/operators.hh>
#include <dune/istl/solvers.hh>
#include <dune/istl/preconditioners.hh>
#include <dune/istl/matrix.hh>

#include "DOFVector.h"
#include "OEMSolver.h"
#include "MemoryManager.h"

namespace AMDiS {
  // ============================================================================
  // ===== class DuneSolver =====================================================
  // ============================================================================

  typedef Dune::BCRSMatrix< Dune::FieldMatrix<double, 1, 1> > DuneMatrix;
  typedef Dune::BlockVector< Dune::FieldVector<double, 1> > DuneVector;
  
  /**
   * \ingroup Solver
   * 
   * \brief
   * Uses ISTL-module from the DUNE project (www.dune-project.com) to solve a
   * linear system of equations.
   */
  template<typename VectorType>
  class DuneSolver : public OEMSolver<VectorType>
  {
  public:
    MEMORY_MANAGED(DuneSolver<VectorType>);

    /** \brief
     * Creator class used in the OEMSolverMap.
     */
    class Creator : public OEMSolverCreator<VectorType>
    {
    public:
      MEMORY_MANAGED(Creator);
      
      virtual ~Creator() {};
      
      /** \brief
       * Returns a new DuneSolver object.
       */
      OEMSolver<VectorType>* create() { 
	return NEW DuneSolver<VectorType>(this->name); 
      };
    };

    /** \brief
     * constructor
     */
    DuneSolver(::std::string name);
    
    /** \brief
     * destructor
     */
    ~DuneSolver();

  protected:
    /** \brief
     * DUNE's solver verbose mode:
     *   0 : print nothing
     *   1 : print initial and final defect and statistics
     *   2 : print line for each iteration
     */
    int verbose_;

    /** \brief
     * DUNE's solver to be used.
     */
    ::std::string solverType_;

    /** \brief
     * DUNE's preconditioner to be used.
     */
    ::std::string preconType_;

    /** \brief
     * Relaxation factor for the preconditioner.
     */
    double preconRelaxation_;

    /** \brief
     * Number of iterations for the preconditioner.
     */
    int preconIterations_;

    /** \brief
     * Implements OEMSolver<VectorType>::init().
     */
    void init() {};
    
    /** \brief
     * Implements OEMSolver<VectorType>::exit().
     */
    void exit() {};

    /** \brief
     * Creates a dune preconditioner using the informations in the used init file. 
     */
    template<typename M, typename V>
    Dune::Preconditioner<V, V>* getPreconditioner(M *matrix);


    /** \brief
     * Creates a dune preconditioner for systems of equations using the informations
     * in the used init file.
     */
    template<typename M, typename V>
    Dune::Preconditioner<V, V>* getSystemPreconditioner(M *matrix);


    /** \brief
     * Creates a dune solver using the informations in the used init file.
     */
    template<typename M, typename V>
    Dune::InverseOperator<V, V>* getSolver(Dune::MatrixAdapter<M, V, V>* ma,
					   Dune::Preconditioner<V, V>* precon);

    /** \brief
     * Implements OEMSolver<VectorType>::solve().
     */
    int solveSystem(MatVecMultiplier<VectorType> *mv, VectorType *x, VectorType *b);

    /** \brief
     * Copies the data of one DOFMatrix to one DuneMatrix. The DuneMatrix has
     * to be allocated before.
     */
    void mapDOFMatrix(DOFMatrix *dofMatrix, DuneMatrix *duneMatrix);

    /** \brief
     * Copies the data of one DOFVector to one DuneVector. The DuneVector has
     * to be allocated before.
     */
    void mapDOFVector(DOFVector<double> *dofVector, DuneVector *duneVector);
  };
}

#include "DuneSolver.hh"

#endif // HAVE_DUNE

#endif // AMDIS_DUNESOLVER_H
