#include "DOFVector.h"

namespace AMDiS
{

  template<typename T>
  void diagPrecond(DOFMatrix *A, DOFVector<T> *vec,
		   DOFVector<BoundaryType> *bound)
  {
    FUNCNAME("QN_Precond::diagPrecond()");

    TEST(A)("No matrix; doing nothing\n");

    DOFMatrix::Iterator rowIterator(A, USED_DOFS);
    typename DOFVector<T>::Iterator vecIterator(vec, USED_DOFS);

    if (bound) {
      DOFVector<BoundaryType>::Iterator
	bIterator(const_cast<DOFVector<BoundaryType>*>(bound), USED_DOFS);
      
      for (vecIterator.reset(), rowIterator.reset(), bIterator.reset();
	   !vecIterator.end(); ++vecIterator, ++bIterator, ++rowIterator)
	{
	  // only non-dirichlet nodes will be preconditioned
	  if ((*bIterator) <= 0 && rowIterator->size() != 0)
	    (*vecIterator) *= (1.0 / (*rowIterator)[0].entry);
	}
    } else {
      for (vecIterator.reset(), rowIterator.reset(); !vecIterator.end();
	   ++vecIterator, ++rowIterator)
	{
	  if (rowIterator->size() != 0)
	    (*vecIterator) *= (1.0 / (*rowIterator)[0].entry);
	}
    }
    
    return;
  }

  template<typename T>
  void aXpY(double alpha, const DOFVector<T>& x, DOFVector<T>& y,
	    DOFVector<BoundaryType> *bound)
  {
    FUNCNAME("QN_Precond::aXpY()");
    const DOFAdmin *admin;

    TEST_EXIT(x.getFESpace() && y.getFESpace())
      ("feSpace is NULL: %8X, %8X\n", x.getFESpace(), y.getFESpace());
    TEST_EXIT((admin = x.getFESpace()->getAdmin()) &&
	      (admin == y.getFESpace()->getAdmin()))
      ("no admin or different admins: %8X, %8X\n",
       x.getFESpace()->getAdmin(), y.getFESpace()->getAdmin());
    TEST_EXIT(static_cast<int>(x.getSize()) >= admin->getUsedSize())
      ("size = %d too small: admin->size = %d\n", x.getSize(),
       admin->getUsedSize());
    TEST_EXIT(static_cast<int>(y.getSize()) >= admin->getUsedSize())
      ("y.size = %d too small: admin->size = %d\n", y.getSize(),
       admin->getUsedSize());

    typename DOFVector<T>::Iterator
      xIterator(dynamic_cast<DOFIndexed<T>*>(const_cast<DOFVector<T>*>(&x)),
		USED_DOFS);
    typename DOFVector<T>::Iterator yIterator(dynamic_cast<DOFIndexed<T>*>(&y),
					      USED_DOFS);

    if (bound) {
      DOFVector<BoundaryType>::Iterator
	bIterator(const_cast<DOFVector<BoundaryType>*>(bound), USED_DOFS);
      
      for (xIterator.reset(), yIterator.reset(), bIterator.reset();
	   !xIterator.end();
	     ++xIterator, ++yIterator, ++bIterator)
	{
	  // only non-dirichlet nodes will be updated
	  if ((*bIterator) <= 0 )
	    *yIterator += alpha * (*xIterator);
	}
    } else {
      for (xIterator.reset(), yIterator.reset(); !xIterator.end();
	   ++xIterator, ++yIterator)
	{
	  *yIterator += alpha * (*xIterator);
	}
    }

    return;
  }

  template<typename T>
  void XpaY(double alpha, const DOFVector<T>& x, DOFVector<T>& y,
	    DOFVector<BoundaryType> *bound)
  {
    FUNCNAME("QN_Precond::aXpY()");
    const DOFAdmin *admin;

    TEST_EXIT(x.getFESpace() && y.getFESpace())
      ("feSpace is NULL: %8X, %8X\n", x.getFESpace(), y.getFESpace());
    TEST_EXIT((admin = x.getFESpace()->getAdmin()) &&
	      (admin == y.getFESpace()->getAdmin()))
      ("no admin or different admins: %8X, %8X\n",
       x.getFESpace()->getAdmin(), y.getFESpace()->getAdmin());
    TEST_EXIT(static_cast<int>(x.getSize()) >= admin->getUsedSize())
      ("size = %d too small: admin->size = %d\n", x.getSize(),
       admin->getUsedSize());
    TEST_EXIT(static_cast<int>(y.getSize()) >= admin->getUsedSize())
      ("y.size = %d too small: admin->size = %d\n", y.getSize(),
       admin->getUsedSize());

    typename DOFVector<T>::Iterator
      xIterator(dynamic_cast<DOFIndexed<T>*>(const_cast<DOFVector<T>*>(&x)),
		USED_DOFS);
    typename DOFVector<T>::Iterator yIterator(dynamic_cast<DOFIndexed<T>*>(&y),
					      USED_DOFS);

    if (bound) {
      DOFVector<BoundaryType>::Iterator
	bIterator(const_cast<DOFVector<BoundaryType>*>(bound), USED_DOFS);
      
      for (xIterator.reset(), yIterator.reset(), bIterator.reset();
	   !xIterator.end();
	   ++xIterator, ++yIterator, ++bIterator)
	{
	  // only non-dirichlet nodes will be updated
	  if ((*bIterator) <= 0 ) {
	    *yIterator *= alpha;
	    *yIterator += *xIterator;
	  }
	}
    } else {
      for (xIterator.reset(), yIterator.reset(); !xIterator.end();
	   ++xIterator, ++yIterator)
	{
	  *yIterator *= alpha;
	  *yIterator += *xIterator;
	}
    }

    return;
  }

}
